/**
 * CPSC 501
 * @author Devon Harstrom built off Jonathan Hudson's starter code
 */

import java.lang.reflect.Constructor;
import java.lang.reflect.Modifier;
import java.lang.reflect.Method;
import java.lang.reflect.Field;
import java.lang.reflect.Array;

public class Inspector {
    public void inspect(Object obj, boolean recursive) {
        Class <?> _class = obj.getClass();
        inspectClass(_class, obj, recursive, 0);
    }

    public void getSuperclass(Class <?> _class, Object obj, boolean recursive, int depth){
        Class <?> _superclass = _class.getSuperclass();
        if (_superclass != null) {
            printIndented(depth, "SUPERCLASS -> Recursively Inspect");
            printIndented(depth, "SuperClass: " + _superclass.getName());
            inspectClass(_superclass, obj, recursive, depth + 1);
        }
        else {
            printIndented(depth, "SuperClass: NONE");
        }
    }

    public void getInterface(  Class <?> _class, Object obj, boolean recursive, int depth){
        Class <?>[] _interface = _class.getInterfaces();
        printIndented(depth, "INTERFACES( " + _class.getName() + " )");

        if(_interface.length == 0) {
            printIndented(depth, "Interfaces-> NONE");
            return;
        }

        printIndented(depth, "Interfaces->");
        for (Class<?> iface : _interface) {
            printIndented(depth, " INTERFACE -> Recursively Inspect");
            printIndented(depth, " " + iface.getName());
            inspectClass(iface, obj, recursive, depth + 1);
        }
    }

    private void printExceptions(int depth, Class<?>[] _exceptions) {
        if(_exceptions.length == 0) {
            printIndented(depth, "  Exceptions-> NONE");
            return;
        }

        printIndented(depth, "  Exceptions->");
        for (Class<?> exception : _exceptions) {
            printIndented(depth, "  class " + exception.getName());
        }
    }

    private void printParameters(int depth, Class <?>[] _parameters){
        if(_parameters.length == 0) {
            printIndented(depth, "  Parameter Types-> NONE");
            return;
        }

        printIndented(depth, "  Parameter types-> ");
        for (Class<?> parameter : _parameters) {
            printIndented(depth, "  " + parameter);
        }
    }

    private void printModifiers(int depth, int int_modifier){
        String _modifier = Modifier.toString(int_modifier);
        printIndented(depth, "  Modifiers: " + _modifier);
    }

    public void getConstructor(  Class <?> _class, Object obj, boolean recursive, int depth) {
        Constructor<?>[] _constructors = _class.getConstructors();
        printIndented(depth, "CONSTRUCTORS( " + _class.getName() + " )");

        if(_constructors.length == 0) {
            printIndented(depth, "Constructors-> NONE");
            return;
        }

        printIndented(depth, "Constructor->");
        for (Constructor<?> constructor : _constructors) {
            printIndented(depth, " CONSTRUCTOR");
            printIndented(depth, "  Name: " + constructor.getName());

            // Exceptions
            Class<?>[] _exceptions = constructor.getExceptionTypes();
            printExceptions(depth, _exceptions);

            //params
            Class<?>[] _parameters = constructor.getParameterTypes(); //parameters
            printParameters(depth, _parameters);

            //modifiers
            int int_modifier = constructor.getModifiers();
            printModifiers(depth, int_modifier);
        }
    }

    public void getMethods(Class <?> _class, Object obj, boolean recursive, int depth){
        printIndented(depth, "METHODS( " + _class.getName() + " )");
        Method[] _methods = _class.getDeclaredMethods();

        if (_methods.length == 0) {
            printIndented(depth, "Methods-> NONE");
            return;
        }

        printIndented(depth, "Methods->");
        for (Method method : _methods) {
            printIndented(depth, " METHOD");
            printIndented(depth, "  Name: " + method.getName());

            // Exceptions
            Class<?>[] _exceptions = method.getExceptionTypes();
            printExceptions(depth, _exceptions);

            //params
            Class<?>[] _parameters = method.getParameterTypes(); //parameters
            printParameters(depth, _parameters);

            // return type
            Class<?> _returnType = method.getReturnType();
            printIndented(depth, "  Return Type: " + _returnType.getName());

            //modifiers
            int int_modifier = method.getModifiers();
            printModifiers(depth, int_modifier);
        }
    }

    private void getFieldValue(Class<?> _type, Object obj, boolean recursive, int depth, Field field){
        field.setAccessible(true);
        try {
            Object _fieldValue = field.get(obj);
            if (_type.isPrimitive()) {
                printIndented(depth, "  Value " + _fieldValue);
                return;
            }
            if (_type.isArray()) {  // Check if it's an array
                Class<?> _componentType = _type.getComponentType();
                printIndented(depth, "  Component Type: " + _componentType.getName());
                int length = Array.getLength(_fieldValue);
                printIndented(depth, "  Length: " + length);
                printIndented(depth, "  Entries->");
                for (int j = 0; j < length; j++) {
                    printIndented(depth, "   Value: " + Array.get(_fieldValue, j));
                }
                return;
            }

            if(_fieldValue == null){
                return;
            }

            printIndented(depth, "  Value (ref): " + _fieldValue);
            if (recursive) {
                printIndented(depth, "   -> Recursively inspect");
                inspectClass(_fieldValue.getClass(), _fieldValue, true, depth + 1);
            }

        } catch (IllegalAccessException e) {
            e.printStackTrace();
        }
    }

    public void getFields(  Class <?> _class, Object obj, boolean recursive, int depth){
        printIndented(depth, "Fields( " + _class.getName() + " )");
        Field[] _fields = _class.getDeclaredFields();

        if(_fields.length == 0) {
            printIndented(depth, "Fields-> NONE");
            return;
        }

        printIndented(depth, "Fields->");
        for (Field field : _fields) {
            printIndented(depth, " FIELD");
            printIndented(depth, "  Name: " + field.getName());

            // type
            Class<?> _type = field.getType();
            printIndented(depth, "  Type:" + _type.getName());

            //modifiers
            int int_modifier = field.getModifiers();
            printModifiers(depth, int_modifier);

            //current value
            getFieldValue(_type, obj, recursive, depth, field);
        }
    }

    private void inspectClass(  Class <?> _class, Object obj, boolean recursive, int depth) {
        printIndented(depth, "CLASS");
        printIndented(depth, "Class: " + _class.getName());
        
        getSuperclass(_class, obj, recursive, depth);
        getInterface(_class, obj, recursive, depth);
        getConstructor(_class, obj, recursive, depth);
        getMethods(_class, obj, recursive, depth);
        getFields(_class, obj, recursive, depth);
    }

    private void printIndented(int depth, String message) {
        System.out.println("\t".repeat(depth) + message);
    }

}


//https://www.baeldung.com/java-check-class-implements-interface#:~:text=The%20getInterfaces()%20Method,implemented%20interfaces%20of%20a%20Class.
// https://docs.oracle.com/javase/8/docs/api/java/lang/reflect/Constructor.html
//https://docs.oracle.com/javase/8/docs/api/java/lang/reflect/package-summary.html
//https://www.tutorialspoint.com/javareflect/javareflect_method_getparametertypes.htm#:~:text=lang.-,reflect.,underlying%20method%20takes%20no%20parameters.
//https://www.geeksforgeeks.org/java-lang-reflect-constructor-class-in-java/
//https://www.geeksforgeeks.org/method-class-getexceptiontypes-method-in-java/
//https://www.geeksforgeeks.org/method-class-getmodifiers-method-in-java/
//https://www.baeldung.com/java-method-reflection
//https://www.geeksforgeeks.org/method-class-getreturntype-method-in-java/
//https://docs.oracle.com/javase/10/docs/api/java/lang/Class.html#getDeclaredFields()
//https://docs.oracle.com/javase/tutorial/reflect/member/fieldTypes.html
//https://www.geeksforgeeks.org/reference-variable-in-java/
//https://www.geeksforgeeks.org/field-get-method-in-java-with-examples/
//https://docs.oracle.com/javase/8/docs/api/java/lang/reflect/Field.html#get-java.lang.Object-
//https://docs.oracle.com/javase/8/docs/api/java/lang/reflect/Array.html
//https://docs.oracle.com/javase/8/docs/api/java/lang/Class.html#getComponentType--

//https://stackoverflow.com/questions/24672749/raw-use-of-parameterized-class